@@ -35,8 +35,9 @@
 #include "crc32c.h"
 #include "extent-cache.h"
 #include "find-root.h"
+#include "help.h"
 
-static void usage(void)
+static void find_root_usage(void)
 {
 	fprintf(stderr, "Usage: find-roots [-a] [-o search_objectid] "
 		"[ -g search_generation ] [ -l search_level ] <device>\n");
@@ -179,19 +180,18 @@ int main(int argc, char **argv)
 			break;
 		case GETOPT_VAL_HELP:
 		default:
-			usage();
+			find_root_usage();
 			exit(c != GETOPT_VAL_HELP);
 		}
 	}
 
 	set_argv0(argv);
-	argc = argc - optind;
-	if (check_argc_min(argc, 1)) {
-		usage();
+	if (check_argc_min(argc - optind, 1)) {
+		find_root_usage();
 		exit(1);
 	}
 
-	fs_info = open_ctree_fs_info(argv[optind], 0, 0,
+	fs_info = open_ctree_fs_info(argv[optind], 0, 0, 0,
 			OPEN_CTREE_CHUNK_ROOT_ONLY |
 			OPEN_CTREE_IGNORE_CHUNK_TREE_ERROR);
 	if (!fs_info) {
@@ -204,8 +204,8 @@ int main(int argc, char **argv)
 			       &filter.match_gen, &filter.match_level);
 	ret = btrfs_find_root_search(fs_info, &filter, &result, &found);
 	if (ret < 0) {
-		fprintf(stderr, "Fail to search the tree root: %s\n",
-			strerror(-ret));
+		errno = -ret;
+		fprintf(stderr, "Fail to search the tree root: %m\n");
 		goto out;
 	}
 	if (ret > 0) {
