@@ -266,11 +266,11 @@ static inline int IS_ERR_OR_NULL(const void *ptr)
 #define div_u64(x, y) ((x) / (y))
 
 /**
- * swap - swap values of @a and @b
+ * __swap - swap values of @a and @b
  * @a: first value
  * @b: second value
  */
-#define swap(a, b) \
+#define __swap(a, b) \
         do { typeof(a) __tmp = (a); (a) = (b); (b) = __tmp; } while (0)
 
 /*
