@@ -1,7 +1,7 @@
 #!/bin/bash
 # Test if mkfs.btrfs --rootdir will skip shrinking correctly
 
-source "$TOP/tests/common"
+source "$TEST_TOP/common"
 
 check_prereq mkfs.btrfs
 
@@ -11,7 +11,7 @@ fs_size=$((512 * 1024 * 1024))
 bs=$((1024 * 1024))
 tmp=$(mktemp -d --tmpdir btrfs-progs-mkfs.rootdirXXXXXXX)
 
-prepare_test_dev $fs_size
+prepare_test_dev "$fs_size"
 
 # No shrink case
 
@@ -20,7 +20,7 @@ run_check_mount_test_dev
 
 # We should be able to write at least half of the fs size data since the fs is
 # not shrunk
-run_check $SUDO_HELPER dd if=/dev/zero bs=$bs count=$(($fs_size / $bs / 2)) \
+run_check $SUDO_HELPER dd if=/dev/zero bs="$bs" count=$(($fs_size / $bs / 2)) \
 	of="$TEST_MNT/file"
 
 run_check_umount_test_dev
@@ -31,7 +31,7 @@ run_check "$TOP/mkfs.btrfs" -f --rootdir "$tmp" --shrink "$TEST_DEV"
 run_check_mount_test_dev
 
 run_mustfail "mkfs.btrfs for shrink rootdir" \
-	$SUDO_HELPER dd if=/dev/zero bs=$bs count=$(($fs_size / $bs / 2)) \
+	$SUDO_HELPER dd if=/dev/zero bs="$bs" count=$(($fs_size / $bs / 2)) \
 	of="$TEST_MNT/file"
 
 run_check_umount_test_dev
